-- top level design to test the bcd to excess-3 converter;
-- to make things "visible" we'll use a clock of 1 Hz to
-- display the result on the on-board LED's; the input is
-- provided via the slide-switches;

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity top_level is
    Port ( reset : in  STD_LOGIC;
           enable : in  STD_LOGIC;
           bcd_par : in  STD_LOGIC_VECTOR (3 downto 0);
           xs3_par : out  STD_LOGIC_VECTOR (3 downto 0);
           clk_100MHz : in  STD_LOGIC);
end top_level;

architecture structural of top_level is

component Code_Converter
  port(	
		enable:	in std_logic;
		X, CLK: in std_logic;
		Z: out std_logic);
end component;

component univ_shift_reg
   generic(N: integer := 4);
   port(
      clk, reset: in std_logic;
      ctrl: in std_logic_vector(1 downto 0);
      d: in std_logic_vector(N-1 downto 0);
      q: out std_logic_vector(N-1 downto 0)
   );
end component;

component ck_divider
    Port (
		CK_IN : in  STD_LOGIC;
		CK_OUT : out  STD_LOGIC);
end component;

type state_t is (S0, S1, S2, S3, S4, S5);
signal state, next_state : state_t;

-- auxiliary signals
signal enable_conv : std_logic;
signal clk_1Hz : std_logic; -- output of clock divider
signal rg_bcd_par : std_logic_vector (3 downto 0);
signal rg_xs3_par : std_logic_vector (3 downto 0);
signal control1 : std_logic_vector (1 downto 0);
signal control2 : std_logic_vector (1 downto 0);

begin

-- instantiate components;
clock : ck_divider port map (
		CK_IN => clk_100MHz,
		CK_OUT => clk_1Hz);	
converter : Code_Converter port map (
		enable => enable_conv,
		X => rg_bcd_par(0),
		CLK => clk_1Hz,
		Z => rg_xs3_par(3));
input_sr : univ_shift_reg 
   generic map (N => 4)
   port map (
      clk => clk_1Hz, 
		reset => reset,
      ctrl => control1,
      d => bcd_par,
      q => rg_bcd_par); -- LSB goes to converter
output_sr : univ_shift_reg 
   generic map (N => 4)
   port map (
      clk => clk_1Hz, 
		reset => reset,
      ctrl => control2,
      d => rg_xs3_par, -- MSB comes from converter
      q => xs3_par);	

-- advance the ad-hoc state machine responsible with 
-- stepping the design thru load parallel and shift serial 
-- to right for 4 clock cycles required by converter to 
-- do its job on 4 bits;
process (reset, enable, clk_1Hz)
begin
	if reset = '1' then
		state <= S0;
	else 
		if enable = '0' then
			state <= S0;
		elsif rising_edge (clk_1Hz) then
			state <= next_state;
		end if;	
	end if;
end process; 

-- state machine responsible with orchestrating the whole operation;
process (enable, clk_1Hz, state)
begin
	case state is
		when S0 =>
			control1 <= "11"; -- load input register;
			control2 <= "00"; -- no op
			enable_conv <= '0';
			if enable = '1' then 
				next_state <= S1;
			else 
				next_state <= S0;
			end if;
		when S1 =>
			control1 <= "10"; -- shift right; one bit processed by converter;
			control2 <= "10"; -- shift right; 
			enable_conv <= '1';
			if enable = '1' then 
				next_state <= S2;
			else 
				next_state <= S0;
			end if;
		when S2 =>
			control1 <= "10"; -- shift right; 2nd bit processed by converter;
			control2 <= "10"; -- shift right;
			enable_conv <= '1';
			if enable = '1' then 
				next_state <= S3;
			else 
				next_state <= S0;
			end if;	
		when S3 => 
			control1 <= "10"; -- shift right; 3rd bit processed by converter;
			control2 <= "10"; -- shift right;
			enable_conv <= '1';
			if enable = '1' then 
				next_state <= S4;
			else 
				next_state <= S0;
			end if;		
		when S4 => 
			control1 <= "10"; -- shift right; 4th bit processed by converter;
			control2 <= "10"; -- shift right;
			enable_conv <= '1';
			if enable = '1' then 
				next_state <= S5;
			else 
				next_state <= S0;
			end if;
		when S5 => 
			control1 <= "00"; -- no op; wait;
			control2 <= "00"; -- no op; wait;
			enable_conv <= '0';
			if enable = '1' then 
				next_state <= S5;
			else 
				next_state <= S0;
			end if;			
	end case;
end process;

end structural;

