-- Greg Stitt
-- University of Florida
--
-- Description: This entity implements a bit difference calculator. Given an
-- input of a generic width, the entity calculates the difference between the
-- number of 1s and 0s. For example, if there are 3 more 1s than 0s, the output is 3.
-- If there are 3 more 0s than 1s, the output is -3.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity bit_diff is
  generic (
    width : positive := 16);
  port (
    clk    : in  std_logic;
    rst    : in  std_logic;
    go     : in  std_logic;
    input  : in  std_logic_vector(width-1 downto 0);
    output : out std_logic_vector(width-1 downto 0);
    done   : out std_logic);
end bit_diff;

architecture FSMD_2P of bit_diff is

  type STATE_TYPE is (S_INIT, S_CHECK_BIT, S_STORE_OUTPUT, S_DONE);

  -- One key difference in the 2-process FSMD model is that anything
  -- implemented as a register must be declared as two signals: one for the
  -- current value and one that defines the register on the next clock cycle.
  -- I use the convention of adding a next_ prefix for the signal that defines
  -- the value on the next clock cycle. Using this convention, anything with a
  -- next_ prefix is implemented with combinational logic, and the
  -- corresponding sigals with the prefix are implemented as registers.
  signal state, next_state : STATE_TYPE;
  signal value, next_value : std_logic_vector(width-1 downto 0);
  signal diff, next_diff   : signed(width-1 downto 0);
  signal count, next_count : integer range 0 to width;

  -- Note the _s suffix on output_s. I use this convention when I need to read
  -- from a signal that is connected to an output. I can't read from "output"
  -- directly because it is an "out" signal, so instead I use an internal
  -- signal with the same name and a _s suffix. I can then read from
  -- "output_s", which is currently assigned to "output". If you are wondering
  -- why the code has to read from "output", the reason is that the
  -- combinational process must define the next value of all registers. For the
  -- output register, the next value will usually be the current value, except
  -- for the S_DONE state. Therefore, we need to be able to read the current
  -- value. 
  signal output_s, next_output : std_logic_vector(width-1 downto 0);
begin

  -- this process defines all registers used in the FSMD
  process(clk, rst)
  begin
    if (rst = '1') then
      value    <= (others => '0');
      count    <= 0;
      diff     <= (others => '0');
      output_s <= (others => '0');
      state    <= S_INIT;
    elsif (clk'event and clk = '1') then
      -- these are the only registers used by the 2-process FSMD
      value    <= next_value;
      count    <= next_count;
      diff     <= next_diff;
      output_s <= next_output;
      state    <= next_state;
    end if;
  end process;

  -- combinational logic
  process(go, input, value, count, diff, output_s, state)
    variable temp : integer range 0 to width;
  begin
    -- assign default values for all combinational outputs. Anything with
    -- the next_ prefix defines the input to a register.
    -- IMPORTANT: Default values are very useful here. If they aren't used,
    -- then each path through the case statement must always define each of
    -- these signals.
    next_count  <= count;
    next_value  <= value;
    next_diff   <= diff;
    next_output <= output_s;
    next_state  <= state;

    -- done isn't a register (although I could have made it one), so we can
    -- assign values to it directly
    done <= '0';

    case state is
      when S_INIT             =>
        -- be careful not to define count, diff, or value, which are assigned
        -- by the sequential process. All register inputs must be assigned
        -- with the signal that has a next_ prefix
        next_count <= 0;
        next_diff  <= (others => '0');
        next_value <= input;

        -- Notice that next_state is not defined for go='0'. Normally, this
        -- would cause a latch, but because I defined a default value for next_
        -- state, no latch is inferred. 
        if (go = '1') then
          next_state <= S_CHECK_BIT;
        end if;

      when S_CHECK_BIT =>

        -- Notice how I'm reading from "diff" and writing to "next_diff".
        -- "diff" is the current register output. "next_diff" is the value
        -- that will be stored in the register on the next cycle.

        if (value(0) = '0') then
          next_diff <= diff - 1;
        elsif (value(0) = '1') then
          next_diff <= diff + 1;
        end if;

        -- Again, I'm assigning "next_value" and not "value".
        next_value <= std_logic_vector(shift_right(unsigned(value), 1));
        temp := count + 1;
        next_count <= temp;

        if (temp = width) then
          next_state <= S_STORE_OUTPUT;
        end if;

        -- In the 2-process model, I had to add an extra state to store the
        -- output. If I had stored the output in the same state that I asserted
        -- done, done would be asserted one cycle before the actual output
        -- appeared, because done is not implemented as a register. If I had
        -- implemented done as a register, I could have eliminated this state.

      when S_STORE_OUTPUT =>
        next_output <= std_logic_vector(diff);
        next_state  <= S_DONE;

      when S_DONE =>
        done       <= '1';
        next_state <= S_INIT;

      when others => null;
    end case;
  end process;

  -- concurrent assignment of internal output_s signal to the corresponding
  -- output. 
  output <= output_s;

end FSMD_2P;
