-- Greg Stitt
-- University of Florida
--
-- Description: This entity implements a bit difference calculator. Given an
-- input of a generic width, the entity calculates the difference between the
-- number of 1s and 0s. If there are 3 more 1s than 0s, the output is 3. If
-- there are 3 more 0s than 1s, the output is -3.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity bit_diff is
  generic (
    width  :     positive := 16);
  port (
    clk    : in  std_logic;
    rst    : in  std_logic;
    go     : in  std_logic;
    input  : in  std_logic_vector(width-1 downto 0);
    output : out std_logic_vector(width-1 downto 0);
    done   : out std_logic);
end bit_diff;

-- The FSM+D model (shown as FSM_D because the + is syntax error)
-- is my recommended model for any time that the
-- exact structure of the datapath is important. As you might have guessed,
-- in this model, you separate the controller and datapath from each other.
-- You then typically define the datapath structurally and then combine it
-- with a corresponding controller (FSM) using any of the possible models.
-- Note that you don't have to remove all datapath operations from the
-- controller. I commonly use hybrid models, where the explicit datapath
-- is used where structure is important (e.g., custom pipelines), while
-- leaving things like counters in the controller as an FSMD.

architecture FSM_D of bit_diff is

  -- control signals between controller (FSM) and datapath
  signal value_ld   : std_logic;
  signal value_sel  : std_logic;
  signal count_ld   : std_logic;
  signal count_sel  : std_logic;
  signal count_done : std_logic;
  signal diff_ld    : std_logic;
  signal diff_sel   : std_logic;
  signal output_ld  : std_logic;
begin

  -- instaniate the controller (FSM) and datapath and connect them together
  -- with the appropriate control signals. See corresponding files datapath.pdf
  -- datapath.vhd, and fsm.vhd.

  U_FSM : entity work.fsm
    port map (
      clk        => clk,
      rst        => rst,
      go         => go,
      done       => done,
      value_ld   => value_ld,
      value_sel  => value_sel,
      count_ld   => count_ld,
      count_sel  => count_sel,
      count_done => count_done,
      diff_ld    => diff_ld,
      diff_sel   => diff_sel,
      output_ld  => output_ld);

  U_DATAPATH : entity work.datapath
    generic map (
      width      => width)
    port map (
      clk        => clk,
      rst        => rst,
      input      => input,
      output     => output,
      value_ld   => value_ld,
      value_sel  => value_sel,
      count_ld   => count_ld,
      count_sel  => count_sel,
      count_done => count_done,
      diff_ld    => diff_ld,
      diff_sel   => diff_sel,
      output_ld  => output_ld);

end FSM_D;
